# AUTO GENERATED ON 2026-02-08 AT 18:23:02
# DO NOT EDIT BY HAND!
#
# To regenerate file, run
#
#     python dev/generate-tests.py
#

# fmt: off

import cupy
import cupy.testing as cpt
import numpy as np
import pytest

import awkward as ak
import awkward._connect.cuda as ak_cu
from awkward._backends.cupy import CupyBackend

cupy_backend = CupyBackend.instance()

@pytest.mark.skip(reason='Unable to generate any tests for kernel')
def test_cudaawkward_RegularArray_combinations_64_1():
    raise NotImplementedError('Unable to generate any tests for kernel')
